/*
 * 	crypto_provider.c
 *
 * Copyright (c) 2004 Evgeniy Polyakov <johnpol@2ka.mipt.ru>
 * 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/err.h>
#include <linux/crypto.h>
#include <linux/mm.h>

#define DEBUG
#include "acrypto.h"
#include "crypto_stat.h"
#include "crypto_def.h"

static void prov_data_ready(struct crypto_device *dev);

static struct crypto_capability prov_caps[] = {
	{CRYPTO_OP_ENCRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_ECB, 1000},
	{CRYPTO_OP_DECRYPT, CRYPTO_TYPE_AES_128, CRYPTO_MODE_ECB, 1000},
};
static int prov_cap_number = sizeof(prov_caps) / sizeof(prov_caps[0]);

static struct completion thread_exited;
static DECLARE_WAIT_QUEUE_HEAD(crypto_wait_queue);
static int need_exit;
static struct crypto_tfm *tfm;

static struct crypto_device pdev = {
	.name = "crypto_provider",
	.data_ready = prov_data_ready,
	.cap = &prov_caps[0],
};

static void prov_data_ready(struct crypto_device *dev)
{
	wake_up(&crypto_wait_queue);
}

static int crypto_thread(void *data)
{
	struct crypto_device *dev = (struct crypto_device *)data;
	u8 *key;
	unsigned int keylen;
	int err, i;
	struct crypto_session *s, *n;
	struct timeval tv[2];

	daemonize("%s", dev->name);
	allow_signal(SIGTERM);

	while (!need_exit) {
		interruptible_sleep_on_timeout(&crypto_wait_queue, 100);

		if (need_exit)
			break;

		list_for_each_entry_safe(s, n, &dev->session_list, dev_queue_entry) {
			if (!spin_trylock(&s->lock))
				continue;

			if (session_completed(s)) {
				goto unlock;
			}
			start_process_session(s);

			do_gettimeofday(&tv[0]);
			dprintk("Begin to process session %llu [%llu] in %s, sg_src_num=%u.\n",
				s->ci.id, s->ci.dev_id, pdev.name, s->data.sg_src_num);

			if (s->data.sg_src_num != s->data.sg_dst_num)
			{
				dprintk("Broken session: sg_src_num [%d] != sg_dst_num [%d].\n",
						s->data.sg_src_num, s->data.sg_dst_num);
				broke_session(s);
				goto out;
			}

			/*
			 * Simple case - key is small(it's size is less than PAGE_SIZE).
			 * Assymetric crypto will require proper key sg handling.
			 */
			key = ((u8 *) page_address(s->data.sg_key[0].page)) + s->data.sg_key[0].offset;
			keylen = s->data.sg_key[0].length;

			err = crypto_cipher_setkey(tfm, key, keylen);
			if (err) {
				dprintk(KERN_ERR "Failed to set key [keylen=%d]: err=%d.\n", keylen, err);
				broke_session(s);
				goto out;
			}

			preempt_count() |= SOFTIRQ_MASK;

			for (i=0; i<s->data.sg_src_num; ++i)
			{
				if (s->ci.operation == CRYPTO_OP_ENCRYPT)
					err = crypto_cipher_encrypt(tfm, &s->data.sg_dst[i], &s->data.sg_src[i], s->data.sg_src[i].length);
				else
					err = crypto_cipher_decrypt(tfm, &s->data.sg_dst[i], &s->data.sg_src[i], s->data.sg_src[i].length);
			
				s->data.sg_dst[i].length = s->data.sg_src[i].length;
				s->data.sg_dst[i].offset = s->data.sg_src[i].offset;
			}
			preempt_count() &= ~SOFTIRQ_MASK;

			do_gettimeofday(&tv[1]);
			dprintk(KERN_INFO "Completing session %llu [%llu] in %s, err=%d, time=%ld.\n",
				s->ci.id, s->ci.dev_id, pdev.name, err,
				(tv[1].tv_sec - tv[0].tv_sec)*1000000 + (tv[1].tv_usec - tv[0].tv_usec));

out:
			crypto_stat_complete_inc(s);
			crypto_session_dequeue_route(s);
			complete_session(s);
			stop_process_session(s);
unlock:
			spin_unlock(&s->lock);
		}
	}

	complete_and_exit(&thread_exited, 0);
}

int prov_init(void)
{
	int err, pid;

	tfm = crypto_alloc_tfm("aes", 0);
	if (!tfm) {
		printk(KERN_ERR "Failed to allocate SHA1 tfm.\n");
		return -EINVAL;
	}

	init_completion(&thread_exited);
	pid = kernel_thread(crypto_thread, &pdev, CLONE_FS | CLONE_FILES);
	if (IS_ERR((void *)pid)) {
		err = -EINVAL;
		dprintk(KERN_ERR "Failed to create kernel load balancing thread.\n");
		goto err_out_free_tfm;
	}

	pdev.cap_number = prov_cap_number;

	err = crypto_device_add(&pdev);
	if (err)
		goto err_out_remove_thread;

	dprintk(KERN_INFO "Test crypto provider module %s is loaded.\n", pdev.name);

	return 0;

err_out_remove_thread:
	need_exit = 1;
	wake_up(&crypto_wait_queue);
	wait_for_completion(&thread_exited);
err_out_free_tfm:
	crypto_free_tfm(tfm);

	return err;
}

void prov_fini(void)
{
	need_exit = 1;
	wake_up(&crypto_wait_queue);
	wait_for_completion(&thread_exited);

	crypto_device_remove(&pdev);
	crypto_free_tfm(tfm);

	dprintk(KERN_INFO "Test crypto provider module %s is unloaded.\n", pdev.name);
}

module_init(prov_init);
module_exit(prov_fini);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Evgeniy Polyakov <johnpol@2ka.mipt.ru>");
MODULE_DESCRIPTION("Test crypto module provider.");
